import uuid

from django.contrib.contenttypes.fields import GenericForeignKey
from django.db import models
from django.utils import timezone

from .signals import payment_received


class Payment(models.Model):
    class StatusType(models.IntegerChoices):
        PENDING = 0, "Ожидает оплаты"
        SUCCESS = 1, "Успешно"
        REJECTED = 2, "Отклонен"

    id = models.UUIDField(
        primary_key=True, unique=True, default=uuid.uuid4, editable=False
    )
    amount = models.DecimalField("Сумма", max_digits=20, decimal_places=2)
    status = models.PositiveSmallIntegerField(
        "Статус", choices=StatusType.choices, default=StatusType.PENDING
    )
    created = models.DateTimeField("Дата создания", default=timezone.now)

    object_id = models.UUIDField()
    content_type = models.ForeignKey(
        "contenttypes.ContentType", on_delete=models.CASCADE
    )
    receiver = GenericForeignKey()

    class Meta:
        verbose_name = "Платеж"
        verbose_name_plural = "Платежи"
        ordering = ["-created"]

    def __str__(self):
        return str(self.id.hex)

    def accept(self):
        self.status = self.StatusType.SUCCESS
        self.save()
        payment_received.send(self.content_type.model_class(), receiver=self.receiver)

    def reject(self):
        self.status = self.StatusType.REJECTED
        self.save()



from django_pay2.exceptions import CreatePaymentError
from django_pay2.decorators import handle_debug
from django_pay2.models import Payment

from .functions import get_payeer_api
from .exceptions import PayeerError


@handle_debug
def create_payeer_payment(request, amount, desc, receiver, currency):
    payment = Payment.objects.create(amount=amount, receiver=receiver)
    try:
        api = get_payeer_api()
        return api.create_payment(str(payment.id), amount, currency, desc)
    except PayeerError as exc:
        payment.reject()
        raise CreatePaymentError(str(exc))



from django.conf import settings
from django.test.signals import setting_changed
from django.urls import reverse_lazy
from django.utils.module_loading import import_string

DEFAULTS = {
    "DEBUG_MODE": False,
    "TEMPLATES": {
        "success": "django_pay2/success_payment.html",
        "rejected": "django_pay2/rejected_payment.html",
    },
    "TINKOFF": {
        "terminal_key": None,
        "password": None,
        "email_company": None,
        "taxation": None,
    },
    "PAYEER": {
        "account": None,
        "api_id": None,
        "api_password": None,
        "shop_id": None,
        "secret_key": None,
    },
    "FREE_KASSA": {
        "merchant_id": None,
        "secret_word_1": None,
        "secret_word_2": None,
    },
    "PERFECT_MONEY": {
        "payee_usd_account": None,
        "payee_eur_account": None,
        "payee_name": None,
        "alternate_passphrase": None,
        "success_url": reverse_lazy("django_pay2:success"),
        "failed_url": reverse_lazy("django_pay2:fail"),
    },
    "COINPAYMENTS": {
        "public_key": None,
        "private_key": None,
        "ipn_secret": None,
        "success_url": reverse_lazy("django_pay2:success"),
        "cancel_url": reverse_lazy("django_pay2:fail"),
    },
    "QIWI": {
        "secret_key": None,
    },
    "QIWI_KASSA": {
        "secret_key": None,
        "site_id": None,
    },
    "SBERBANK": {
        "username": None,
        "password": None,
        "return_url": reverse_lazy("django_pay2:success"),
        "fail_url": reverse_lazy("django_pay2:fail"),
        "public_key": None,
    },
}

IMPORT_DICTS = []


class ObjDict(dict):
    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError:
            raise AttributeError(f"Invalid setting: {item}")


class ImportDict(ObjDict):
    def __getattr__(self, item):
        val = super().__getattr__(item)
        if isinstance(val, str):
            return import_string(val)
        elif isinstance(val, (list, tuple)):
            return [import_string(s) for s in val]
        else:
            raise ValueError(f"Setting must be a string or list: {item}")


class PaymentSettings:
    def __init__(self, user_settings=None, defaults=None, import_dicts=None):
        if user_settings:
            self._user_settings = user_settings
        self.defaults = defaults or DEFAULTS
        self.import_dicts = import_dicts or IMPORT_DICTS
        self._cached_attrs = set()

    @property
    def user_settings(self):
        if not hasattr(self, "_user_settings"):
            self._user_settings = getattr(settings, "PAYMENTS", {})
        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError(f'Invalid payment settings: "{attr}"')

        try:
            val = self.user_settings[attr]
        except KeyError:
            val = self.defaults[attr]

        if isinstance(val, dict):
            dict_class = ImportDict if attr in self.import_dicts else ObjDict
            default = dict_class(self.defaults[attr])
            default.update(dict_class(val))
            val = default

        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, "_user_settings"):
            delattr(self, "_user_settings")


payment_settings = PaymentSettings()


def reload_app_settings(*args, **kwargs):
    setting = kwargs["setting"]
    if setting == "PAYMENTS":
        payment_settings.reload()


setting_changed.connect(reload_app_settings)



import logging

from rest_framework import generics
from rest_framework.response import Response

from .serializers import QiwiKassaNotifySerializer
from .contants import NotifyType, Status

logger = logging.getLogger(__name__)


class QiwiKassaNotifyView(generics.GenericAPIView):
    def post(self, request, *args, **kwargs):
        notify_type = request.data.get("type")
        if notify_type != NotifyType.PAYMENT:
            msg = (
                f"Qiwi Kassa notify with type={notify_type} cannot be processed. "
                "Just ignoring it."
            )
            logger.info(msg)
            return Response({"msg": msg})

        request_hmac = request.meta.get("HTTP_X_API_SIGNATURE_SHA256")
        serializer = QiwiKassaNotifySerializer(
            data=request.data,
            context={
                "hmac": request_hmac,
            },
        )
        if not serializer.is_valid():
            logger.info(
                f"Error with processing request.data={request.data} hmac={request_hmac}"
            )
            serializer.is_valid(raise_exception=True)

        status = serializer.validated_data["payment"]["status"]["value"]
        if status != Status.SUCCESS:
            msg = (
                f"Qiwi Kassa notify with status={status} cannot be processed. "
                "Just ignoring it."
            )
            logger.info(msg)
            return Response({"msg": msg})

        payment = serializer.validated_data["payment"]["bill_id"]
        payment.accept()

        return Response({"msg": "ok"})



import functools

from django.urls import reverse

from .models import Payment
from .settings import payment_settings
from .payment_methods import PaymentForm, PaymentRedirect


def handle_debug(func):
    @functools.wraps(func)
    def wrapped(request, amount, desc, receiver, *args, **kwargs):
        if payment_settings.DEBUG_MODE:
            payment = Payment.objects.create(
                amount=amount,
                receiver=receiver,
            )
            return PaymentRedirect(
                url=reverse("django_pay2:debug_payment", args=[payment.id])
            )
        return func(request, amount, desc, receiver, *args, **kwargs)

    return wrapped


def handle_form_debug(func):
    @functools.wraps(func)
    def wrapped(request, amount, desc, receiver, *args, **kwargs):
        if payment_settings.DEBUG_MODE:
            payment = Payment.objects.create(
                amount=amount,
                receiver=receiver,
            )
            return PaymentForm(
                action=reverse("django_pay2:debug_payment", args=[payment.id]),
                fields={"amount": str(amount), "desc": desc},
            )
        return func(request, amount, desc, receiver, *args, **kwargs)

    return wrapped



# flake8: noqa
from .api import TinkoffApi, TinkoffOrderItem
from .constants import (
    Languages,
    PayTypes,
    Taxations,
    PaymentMethods,
    Taxes,
    PaymentObjects,
)
from .exceptions import TinkoffApiError, TinkoffNetworkError, TinkoffResponseError



import logging

from rest_framework import serializers

from django_pay2.models import Payment
from django_pay2.utils import HmacDefault, MappingFieldsMixin

from .api import get_api

logger = logging.getLogger(__name__)


class StatusSerializer(serializers.Serializer):
    value = serializers.CharField()


class AmountSerializer(serializers.Serializer):
    value = serializers.DecimalField(max_digits=20, decimal_places=2)
    currency = serializers.CharField()


class QiwiPaymentSerializer(MappingFieldsMixin, serializers.Serializer):
    mapping_keys = {
        "payment_id": "paymentId",
        "bill_id": "billId",
        "created_datetime": "createdDateTime",
    }

    payment_id = serializers.CharField()
    bill_id = serializers.PrimaryKeyRelatedField(queryset=Payment.objects.all())
    amount = AmountSerializer()
    status = StatusSerializer()
    created_datetime = serializers.CharField()

    default_error_messages = {"mismatched_amount": "Mismatched amount"}

    def validate(self, attrs):
        attrs = super().validate(attrs)
        bill = attrs["bill_id"]
        if bill.amount != attrs["amount"]["value"]:
            self.fail("mismatched_amount")
        return attrs


class QiwiKassaNotifySerializer(serializers.Serializer):
    payment = QiwiPaymentSerializer()
    hmac = serializers.HiddenField(default=HmacDefault())

    default_error_messages = {"mismatch_hmac": "Mismatch HMAC"}

    def validate(self, attrs):
        attrs = super().validate(attrs)
        expected_hmac = get_api().create_hmac(
            payment_id=attrs["payment"]["payment_id"],
            created_datetime=attrs["payment"]["created_datetime"],
            amount=attrs["payment"]["amount"]["value"],
        )
        if expected_hmac != attrs["hmac"]:
            logger.info(f"Mismatch hmac: expected={expected_hmac} got={attrs['hmac']}")
            self.fail("mismatch_hmac")
        return attrs



import requests
import decimal
import hashlib

from django_pay2.models import Payment
from .exceptions import PayeerError, PayeerValidationError, AlreadyPaid
from django_pay2.payment_methods import PaymentRedirect


class PayeerApi:
    url = "https://payeer.com/ajax/api/api.php"

    def __init__(
        self,
        account: str,
        api_id: str,
        api_password: str,
        secret_key: str,
        shop_id: str,
    ):
        self.account = account
        self.api_id = (api_id,)
        self.api_password = api_password
        self.secret_key = secret_key
        self.shop_id = shop_id

    def create_payment(
        self, order_id: str, amount: decimal.Decimal, currency: str, desc: str
    ) -> str:
        response = requests.post(
            self.url,
            {
                "account": self.account,
                "apiId": self.api_id,
                "apiPass": self.api_password,
                "action": "invoiceCreate",
                "m_shop": self.shop_id,
                "m_orderid": order_id,
                "m_amount": float(amount),
                "m_curr": currency,
                "m_desc": desc,
            },
        )
        response.raise_for_status()
        response_data = response.json()
        if not response_data["success"]:
            raise PayeerError(response_data["errors"])
        return PaymentRedirect(response_data["url"])

    def notify(self, data):
        try:
            request = PayeerNotifyRequest(data)
        except (ValueError, TypeError, decimal.ConversionSyntax):
            raise PayeerValidationError("Invalid data format")
        try:
            payment = Payment.objects.get(id=request.order_id)
        except Payment.DoesNotExist:
            raise PayeerValidationError("Payment does not exist", request.order_id)
        if payment.status != Payment.StatusType.PENDING:
            raise AlreadyPaid(request.order_id)
        if payment.amount != request.amount:
            raise PayeerValidationError("Amounts are not equal", request.order_id)
        if request.sign != self._calculate_sign(request):
            raise PayeerValidationError("Incorrect sign", request.order_id)
        return PayeerNotifyResult(payment, request.status, request.order_id)

    def _calculate_sign(self, notify_request: "PayeerNotifyRequest"):
        hash_list = [
            notify_request.operation_id,
            notify_request.operation_ps,
            notify_request.operation_date,
            notify_request.operation_pay_date,
            notify_request.shop,
            notify_request.order_id,
            notify_request.amount,
            notify_request.currency,
            notify_request.desc,
            notify_request.status,
            self.secret_key,
        ]
        hash_list = [str(s) for s in hash_list]
        return hashlib.sha256(":".join(hash_list).encode("utf-8")).hexdigest().upper()


class PayeerNotifyRequest:
    def __init__(self, raw_data: dict):
        self.operation_id = raw_data.get("m_operation_id")
        self.operation_ps = raw_data.get("m_operation_ps")
        self.operation_date = raw_data.get("m_operation_date")
        self.operation_pay_date = raw_data.get("m_operation_pay_date")
        self.shop = raw_data.get("m_shop")
        self.order_id = raw_data.get("m_orderid")
        self.amount = decimal.Decimal(raw_data.get("m_amount"))
        self.currency = raw_data.get("m_curr")
        self.desc = raw_data.get("m_desc")
        self.status = raw_data.get("m_status")
        self.sign = raw_data.get("m_sign")


class PayeerNotifyResult:
    def __init__(self, payment, status, raw_order_id):
        self.payment = payment
        self.status = status
        self.raw_order_id = raw_order_id



import logging

from rest_framework.views import APIView
from rest_framework import status, renderers
from rest_framework.response import Response

from .serializers import NotifySerializer

logger = logging.getLogger(__file__)


class NotifyView(APIView):
    renderer_classes = (renderers.StaticHTMLRenderer,)

    def post(self, request, *args, **kwargs):
        serializer = NotifySerializer(
            data=request.data, context={"raw_data": request.data}
        )
        if not serializer.is_valid():
            logger.warning(
                f"Tinkoff notification error: {serializer.errors}, orig data: {request.data}"
            )
            return Response(status=status.HTTP_400_BAD_REQUEST)
        serializer.validated_data["payment"].accept()
        return Response("OK")



from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework import status, parsers

from .serializers import PerfectMoneyNotifySerializer


class PerfectMoneyNotifyView(APIView):
    parser_classes = (parsers.FormParser,)

    def post(self, request, *args, **kwargs):
        serializer = PerfectMoneyNotifySerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        payment = serializer.validated_data["payment_id"]
        payment.accept()
        return Response(status=status.HTTP_204_NO_CONTENT)



from rest_framework import serializers

from django_pay2.models import Payment

from .api import get_api


class PerfectMoneyNotifySerializer(serializers.Serializer):
    payee_account = serializers.CharField()
    payment_id = serializers.PrimaryKeyRelatedField(queryset=Payment.objects.all())
    payment_amount = serializers.DecimalField(decimal_places=2, max_digits=20)
    payment_units = serializers.CharField()
    payment_batch_num = serializers.CharField()
    payer_account = serializers.CharField()
    timestamp = serializers.CharField()
    sign = serializers.CharField()

    default_error_messages = {
        "incorrect_sign": "Incorrect sign",
        "mismatch_payment": "Mismatch payment",
        "incorrect_amount": "Incorrect amount",
    }

    def to_internal_value(self, data):
        new_data = {
            "payee_account": data.get("PAYEE_ACCOUNT"),
            "payment_id": data.get("PAYMENT_ID"),
            "payment_amount": data.get("PAYMENT_AMOUNT"),
            "payment_units": data.get("PAYMENT_UNITS"),
            "payment_batch_num": data.get("PAYMENT_BATCH_NUM"),
            "payer_account": data.get("PAYER_ACCOUNT"),
            "timestamp": data.get("TIMESTAMPGMT"),
            "sign": data.get("V2_HASH"),
        }
        return super().to_internal_value(new_data)

    def validate(self, attrs):
        if attrs["payment_amount"] != attrs["payment_id"].amount:
            self.fail("incorrect_amount")

        api = get_api()
        expected_sign = api.calculate_sign(
            attrs["payment_id"].id,
            attrs["payment_amount"],
            attrs["payment_units"],
            attrs["payment_batch_num"],
            attrs["payer_account"],
            attrs["timestamp"],
        )
        if expected_sign != attrs["sign"]:
            self.fail("incorrect_sign")

        return attrs



from rest_framework import serializers

from django_pay2.models import Payment
from django_pay2.settings import payment_settings

from .api import get_api


class FreeKassaNotifySerializer(serializers.Serializer):
    merchant_id = serializers.CharField()
    amount = serializers.FloatField()
    order = serializers.PrimaryKeyRelatedField(queryset=Payment.objects.all())
    sign = serializers.CharField()

    default_error_messages = {
        "incorrect_merchant_id": "Incorrect merchant ID",
        "amounts_not_equal": "Amounts are not equal",
        "incorrect_sign": "Incorrect sign",
    }

    def validate_merchant_id(self, value):
        if payment_settings.FREE_KASSA.merchant_id != value:
            self.fail("incorrect_merchant_id")

    def validate(self, attrs):
        attrs = super().validate(attrs)
        if attrs["order"].amount != attrs["amount"]:
            self.fail("amounts_not_equal")
        api = get_api()
        expected_sign = api.generate_notify_sign(attrs["amount"], attrs["order"].id)
        if expected_sign != attrs["sign"]:
            self.fail("incorrect_sign")
        return attrs



import logging

from django.views.generic import View
from django.http import HttpResponse

from .serializers import SberbankCallbackSerializer
from .constants import SberbankCallbackStatus, SberbankCallbackOperation

logger = logging.getLogger(__file__)


class CallbackView(View):
    def get(self, request):
        serializer = SberbankCallbackSerializer(data=request.GET.dict())
        if not serializer.is_valid():
            logger.info(
                f"Validation error for sberbank callback. "
                f"Data: {serializer.initial_data}. "
                f"Errors: {serializer.errors}"
            )
            return HttpResponse(status=400)

        payment = serializer.validated_data["order_num"]

        operation = serializer.validated_data["operation"]
        if operation != SberbankCallbackOperation.approved:
            logger.info(
                f'Got callback for payment "{payment.id}": operation {operation}'
            )
            return HttpResponse(status=200)

        status = serializer.validated_data["status"]
        if status != SberbankCallbackStatus.success:
            logger.info(
                f'Got callback for payment "{payment.id}": '
                f"operation {operation} finished with error"
            )
            return HttpResponse(status=200)

        payment.accept()
        logger.info(f'Got callback for payment "{payment.id}": accepted.')
        return HttpResponse(status=200)



from django.views import generic
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.http import QueryDict, HttpResponse

from .serializers import CoinPaymentsApproveSerializer
from .api import get_api


@method_decorator(csrf_exempt, "dispatch")
class CoinPaymentsNotifyView(generic.View):
    def post(self, request, *args, **kwargs):
        body = request.body
        encoding = request.encoding if request.encoding is not None else "utf-8"
        decoded_body = body.decode(encoding)
        expected_hmac = get_api().calculate_ipn_hmac(decoded_body)
        if request.META.get("HTTP_HMAC") != expected_hmac:
            return self.ipn_error("HMAC signatures do not match")

        data = QueryDict(decoded_body)
        serializer = CoinPaymentsApproveSerializer(data=data)
        if not serializer.is_valid():
            return self.ipn_error("Validation error")
        payment = serializer.validated_data["invoice"]
        payment.accept()
        return HttpResponse("IPN OK")

    def ipn_error(self, message):
        return HttpResponse(f"IPN Error: {message}")



from django.db.models import TextChoices


class NotifyType(TextChoices):
    PAYMENT = "PAYMENT"
    CAPTURE = "CAPTURE"
    REFUND = "REFUND"
    CHECK_CARD = "CHECK_CARD"


class Status(TextChoices):
    SUCCESS = "SUCCESS"
